/*
  process.cpp
  Copyright (C) 2010 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of radiko-player.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#include "process.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <qstring.h>
#include <sys/select.h>
#include <unistd.h>
#include <qapplication.h>

FILE *fp;

void dump(const void *buf, int size)
{
	int i;
	const char *p = (const char *)buf;
	for (i = 0; i < size; i++) {
		printf("%02x ", *p++);
		if (i % 16 == 15)
			printf("\n");
	}
	printf("\n");
}


Process::Process(QString cmd)
	: timer(this)
{
	run = false;
	this->cmd = cmd;
	connect(&timer, SIGNAL(timeout()),
			this, SLOT(check_process()));

#ifdef QWS
	fp = fopen("/dev/ttya0", "w");
#else
	fp = stdout;
#endif
}

Process::~Process()
{
	printf("%s enter\n", __PRETTY_FUNCTION__);
	if (run) {
		stop();
		wait();
	}
	printf("%s end\n", __PRETTY_FUNCTION__);
}

void Process::setCommand(QString cmd)
{
	this->cmd = cmd;
}

void Process::start()
{
	int p;
	int p_stdin[2];
	int p_stdout[2];
	int p_stderr[2];

	printf("start\n");

	if (run)
		stop();

	if (pipe(p_stdin))
		return;

	if (pipe(p_stdout))
		return;

	if (pipe(p_stderr))
		return;

	if ((p = fork()) == 0) {
		/* child */
		int ret;
		char command[1024];
		int i;
//		FILE *s_in, *s_out, *s_err;

		printf("write\n");

		/* duplicate fd */
		dup2(p_stdin[0], 0);
		dup2(p_stdout[1], 1);
		dup2(p_stderr[1], 2);

		/* close other file descriptors */
		for (i = 3; i < 256; i++)
			close(i);

		setpgid(0, 0);
//		fprintf(s_out, "c getpgrp = %d\n", getpgrp());

		strcpy(command, cmd.latin1());
//		strcpy(command, "echo hogehoge > /dev/ttya0");

//		printf("cmd: %s\n", command);

		ret = system(command);
//		ret = execl("/bin/sh", "sh", "-c", command, NULL);
//		fprintf(stderr, "ret:%d\n", ret);
		exit(ret);
		/* no return */
	}

	/* parent */
	pid = p;
	fprintf(fp, "pid: %d\n", pid);

	printf("in:%d-%d out:%d-%d err:%d-%d\n",
		   p_stdin[0], p_stdin[1], p_stdout[0], p_stdout[1],
		   p_stderr[0],p_stderr[1]);

	close(p_stdin[0]);
	close(p_stdout[1]);
	close(p_stderr[1]);

#if 0
	char buf[100];
	printf("re\n");
	int ret = read(p_stdout[0], buf, 100);
	printf("print %s (ret=%d)\n", buf, ret);
#endif
	fd_stdin = p_stdin[1];
	fd_stdout = p_stdout[0];
	fd_stderr = p_stderr[0];
	f_stdin = fdopen(fd_stdin, "w");
	f_stdout = fdopen(fd_stdout, "r");
	f_stderr = fdopen(fd_stderr, "r");


#if 0
	c_stdin.open(p_stdin[1], IO_WriteOnly);
	c_stdout.open(p_stdout[0], IO_ReadOnly);
	c_stderr.open(p_stderr[0], IO_ReadOnly);

	char b[100];
	printf("read\n");
	c_stdout.readBlock(b, 8);
	printf("%s", b);
#endif

	timer.start(100);

	emit started();
	run = true;
}

void Process::stop()
{
	if (!run)
		return;

	killpg(pid, 9);
//	system("killall -9 mplayer");
//	run = false;
//	timer.stop();
//	emit exited(0);
}

void Process::wait()
{
	if (!run)
		return;

//	printf("wait child pid = %d\n", pid);
	while (run) {
		check_process();
		qApp->processOneEvent();
	}
}

void Process::check_process()
{
	int ret;
	char buf[256];

	ret = waitpid(pid, &status, WNOHANG);
	//	fprintf(fp, "check_process pid=%d ret=%d\n", pid, ret);
	if (ret != 0) {
		run = false;
		timer.stop();
		fprintf(fp, "check_process: exit child pid = %d status = %d\n", pid, status);

		c_stdin.close();
		c_stdout.close();
		c_stderr.close();
		timer.stop();

//		emit exited(status);
		emit exited();
		return;
	}

//	printf("buffer out:%d err:%d\n",
//		   isBuffered(fd_stdout), isBuffered(fd_stderr));

	while (isBuffered(fd_stdout)) {
		if (!fgets(buf, sizeof(buf) - 1, f_stdout))
			break;
//		dump(buf, strlen(buf));
		emit recvStdout(QString(buf));
//		printf("stdout:\n%s\n", buf);
	}

	while (isBuffered(fd_stderr)) {
		if (!fgets(buf, sizeof(buf) - 1, f_stderr))
			break;
		emit recvStderr(QString(buf));
//		printf("stdout:\n%s\n", buf);
	}

	fflush(fp);

}

bool Process::isBuffered(int fd)
{
	struct timeval tv = {0, 0};
	fd_set rfds;
	int ret;

	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);

	ret = select(fd + 1, &rfds, NULL, NULL, &tv);
	return (ret == 1);
}

void Process::sendStdin(const QString msg)
{
//	printf("%s\n", __PRETTY_FUNCTION__);
	if (!run)
		return;

//	printf("stdin:%s", msg.latin1());
	fputs(msg.latin1(), f_stdin);
	fflush(f_stdin);
}
